import pytest
import torch
from botorch.models import SingleTaskGP

from tests.acquisition.helpers import check_posterior, check_same_model_data


def get_train_x():
    gen = torch.Generator()
    _ = gen.manual_seed(7)
    return torch.rand(10, 1, dtype=torch.float64, generator=gen)


# lets just write wht whole test without importing any of our stuff for now


# This warning is due to train_y not being standardised. Standarisation can improve quality of fit.
# This is not imporant for this test.
@pytest.mark.filterwarnings(
    "ignore : Input data is not standardized : botorch.exceptions.InputDataWarning : botorch.models"
)
@pytest.mark.external
@pytest.mark.parametrize(
    "n_targets, fixed_noise",
    [
        pytest.param(1, False, id="Single target, homoskedatic noise"),
        pytest.param(2, False, id="Multiple target, homoskedatic noise"),
        pytest.param(1, True, id="Single target, Fixed noise"),
        pytest.param(2, True, id="Multiple target, Fixed noise"),
    ],
)
def test_condition_on_observations(n_targets: int, fixed_noise: bool):  # noqa: FBT001
    """Checks if the conditional/fantasy produces the same model as training with all the data.

    Args:
        n_targets: The number of targets the GP should predict.
        fixed_noise: If True, use fixed noise GP, other wise use infered homoskedastic noise.

    Note: This is possible because `SingleTaskGP` does not fit the hyper parmas, that is done using
    >>> mll = gpytorch.mlls.ExactMarginalLogLikelihood(model.likelihood, model)
    >>> fit_gpytorch_model(mll)

    Benefits of using condition_on_observations:
    - Input have consistent shape, and match instationation of SingleTaskGp.

    Gotchas:
    - some of the arguments are expected to be in model space, others in output (problem) space. Details https://github.com/pytorch/botorch/issues/2533
    """
    fixed_noise = True
    n_targets = 1

    train_x = get_train_x()
    n = train_x.shape[0]
    train_y = torch.sin(train_x).expand(n, n_targets)
    train_yvar = torch.ones_like(train_y) * 0.1

    model_original = SingleTaskGP(train_x, train_y, train_yvar if fixed_noise else None)

    # run the model once to avoid runtime error detailing this requirement
    _ = model_original.posterior(torch.tensor([[0.5]]))

    # Generate new training data to use
    new_x = torch.tensor([[0.5]])  # (n,d)
    new_y = torch.tensor([[0.1]]).expand(1, n_targets)  # (n, m)
    new_yvar = torch.ones_like(new_y) * 0.1

    # New conditioned model
    model_conditioned = model_original.condition_on_observations(
        # MUST BE IN MODEL SPACE - will not be automatically transformed
        # shape require (b,n',d)
        X=new_x,
        # Must be in outcome space - will be automatically transformed
        # Shape require (b,n',m)
        Y=new_y,
        # Must be in model space - - will not be automatically transformed
        noise=new_yvar if fixed_noise else None,
    )

    # Expected model
    x_total = torch.vstack([train_x, new_x])  # (n+1,d=1)
    y_total = torch.vstack([train_y, new_y])  # (n+1,n_targets)
    yvar_total = torch.vstack([train_yvar, new_yvar])  # (n+1,n_targets)
    model_expected = SingleTaskGP(x_total, y_total, yvar_total if fixed_noise else None)

    # Check the internal data is the same
    check_same_model_data(model_actual=model_conditioned, model_expected=model_expected)
    # check that they both make the same preditiontion
    x = torch.tensor([[0.4], [0.7]])
    expected_posterior = model_expected.posterior(x)
    conditioned_posterior = model_conditioned.posterior(x)
    check_posterior(conditioned_posterior, expected_posterior)


@pytest.mark.external
# This warning is due to train_y not being standardised. Standarisation can improve quality of fit.
# This is not imporant for this test.
@pytest.mark.filterwarnings(
    "ignore : Input data is not standardized : botorch.exceptions.InputDataWarning : botorch.models"
)
@pytest.mark.xfail(
    reason="Batched GP has been created from non-batch GP",
)
def test_condition_on_observations_additional_data_incorrect_dim():
    """Checks the outcome of the model if adding batched data to a non-batched model.

    OUTCOME: new model hase shape torch.Size([1, 11, 1]), rather than the expected torch.Size([11, 1]).
    This because the model has now been turned into a batched GP
    """
    train_x = get_train_x()
    train_y = torch.sin(train_x)
    train_yvar = torch.ones_like(train_y) * 0.1

    model_original = SingleTaskGP(train_x, train_y, train_yvar)

    # run the model once to avoid runtime error detailing this requirement
    _ = model_original.posterior(torch.tensor([[0.5]]))

    # Make the new data have a batch dimension. If the original model was trained with data of this dimension it would
    # be a batched model (see here: https://botorch.org/docs/batching#batched-evaluation-of-models-and-acquisition-functions)
    new_x = torch.tensor([[[0.5]]])  # (b,n,d)
    new_y = torch.tensor([[[0.1]]])  # (b,n, m)
    new_yvar = torch.ones_like(new_y) * 0.1

    # New conditioned model
    model_conditioned = model_original.condition_on_observations(
        # MUST BE IN MODEL SPACE - will not be automatically transformed
        # shape require (b,n',d)
        X=new_x,
        # Must be in outcome space - will be automatically transformed
        # Shape require (b,n',m)
        Y=new_y,
        # Must be in model space - - will not be automatically transformed
        noise=new_yvar,
    )

    # Expected model
    x_total = torch.vstack([train_x, new_x[0]])  # (n+1,d=1)
    y_total = torch.vstack([train_y, new_y[0]])  # (n+1,n_targets)
    yvar_total = torch.vstack([train_yvar, new_yvar[0]])  # (n+1,n_targets)
    model_expected = SingleTaskGP(x_total, y_total, yvar_total)

    # Check the internal data is the same
    check_same_model_data(model_actual=model_conditioned, model_expected=model_expected)
    # check that they both make the same preditiontion
    x = torch.tensor([[0.4], [0.7]])
    expected_posterior = model_expected.posterior(x)
    conditioned_posterior = model_conditioned.posterior(x)
    check_posterior(conditioned_posterior, expected_posterior)


@pytest.mark.external
# This warning is due to train_y not being standardised. Standarisation can improve quality of fit.
# This is not imporant for this test.
@pytest.mark.filterwarnings(
    "ignore : Input data is not standardized : botorch.exceptions.InputDataWarning : botorch.models"
)
@pytest.mark.xfail(
    reason="Supplying noise in to update homoskedatic has side effects",
)
def test_condition_on_observations_silent_failure():
    """Demonstate the silent failure when providing noise to a GP that doesn't use it.

    The posteriors produces are not the same. This is an edge case of `test_condition_on_observations`.
    """
    fixed_noise = False
    n_targets = 1

    train_x = get_train_x()
    n = train_x.shape[0]
    train_y = torch.sin(train_x).expand(n, n_targets)
    train_yvar = torch.ones_like(train_y) * 0.1

    model_original = SingleTaskGP(train_x, train_y, train_yvar if fixed_noise else None)

    # run the model once to avoid runtime error detailing this requirement
    _ = model_original.posterior(torch.tensor([[0.5]]))

    # Generate new training data to use
    new_x = torch.tensor([[0.5]])  # (n,d)
    new_y = torch.tensor([[0.1]]).expand(1, n_targets)  # (n, m)
    new_yvar = torch.ones_like(new_y) * 0.1

    # New conditioned model
    model_conditioned = model_original.condition_on_observations(
        # MUST BE IN MODEL SPACE - will not be automatically transformed
        # shape require (b,n',d)
        X=new_x,
        # Must be in outcome space - will be automatically transformed
        # Shape require (b,n',m)
        Y=new_y,
        # Must be in model space - - will not be automatically transformed
        noise=new_yvar,
    )

    # Expected model
    x_total = torch.vstack([train_x, new_x])  # (n+1,d=1)
    y_total = torch.vstack([train_y, new_y])  # (n+1,n_targets)
    yvar_total = torch.vstack([train_yvar, new_yvar])  # (n+1,n_targets)
    model_expected = SingleTaskGP(x_total, y_total, yvar_total if fixed_noise else None)

    # Check the internal data is the same
    check_same_model_data(model_actual=model_conditioned, model_expected=model_expected)
    # check that they both make the same preditiontion
    x = torch.tensor([[0.4], [0.7]])
    expected_posterior = model_expected.posterior(x)
    conditioned_posterior = model_conditioned.posterior(x)

    check_posterior(conditioned_posterior, expected_posterior)


# This warning is due to train_y not being standardised. Standarisation can improve quality of fit.
# This is not imporant for this test.
@pytest.mark.filterwarnings(
    "ignore : Input data is not standardized : botorch.exceptions.InputDataWarning : botorch.models"
)
@pytest.mark.external
@pytest.mark.parametrize(
    "n_targets, fixed_noise",
    [
        pytest.param(1, False, id="Single target, homoskedatic noise"),
        pytest.param(2, False, id="Multiple target, homoskedatic noise"),
        pytest.param(1, True, id="Single target, Fixed noise"),
        pytest.param(2, True, id="Multiple target, Fixed noise"),
    ],
)
def test_get_fantasy_model(n_targets: int, fixed_noise: bool):  # noqa: FBT001
    """Checks if the conditional/fantasy produces the same model as training with all the data.

    Args:
        n_targets: The number of targets the GP should predict.
        fixed_noise: If True, use fixed noise GP, other wise use infered homoskedastic noise.

    Note: This is possible because `SingleTaskGP` does not fit the hyper parmas, that is done using
    >>> mll = gpytorch.mlls.ExactMarginalLogLikelihood(model.likelihood, model)
    >>> fit_gpytorch_model(mll)

    Cons:
    - Dimesnion changes in shape
    - Less clear documentation

    Gotchas:
    - everything needs to be in model-space
    """
    train_x = get_train_x()
    n = train_x.shape[0]
    train_y = torch.sin(train_x).expand(n, n_targets)
    train_yvar = torch.ones_like(train_y) * 0.1

    model_original = SingleTaskGP(train_x, train_y, train_yvar if fixed_noise else None)

    # run the model once to avoid runtime error detailing this requirement
    _ = model_original.posterior(torch.tensor([[0.5]]))

    # Generate new training data to use
    new_x = torch.tensor([[0.5]])  # (n,d)
    new_y = torch.tensor([[0.1]]).expand(1, n_targets)  # (n, m)
    new_yvar = torch.ones_like(new_y) * 0.1

    # reshape the new_y and new_yvar to match the format of the underling model. This is required for this method
    # gpytorch docs say: (f,*b, m):
    # In botorch is is written as (*b,n') - which is not super clear because there is no m (target)
    # In reality this just need to match the dimensions of model.train_targets.
    # (n_targets,n) when n_targets> 1
    # (n,) when n_targets==1
    reshaped_new_y = new_y.reshape(-1) if n_targets == 1 else new_y.T
    reshaped_new_yvar = torch.ones_like(reshaped_new_y) * 0.1

    if fixed_noise:
        model_conditioned = model_original.get_fantasy_model(
            # gpytorch docs say: (f,*b, m, d). In botorch is is written as (*b,n',d)
            inputs=new_x,
            targets=reshaped_new_y,
            # This is Kwarg that gets passed through.  must match new_y
            noise=reshaped_new_yvar,
        )
    else:
        model_conditioned = model_original.get_fantasy_model(
            # gpytorch docs say: (f,*b, m, d). In botorch is is written as (*b,n',d)
            inputs=new_x,
            targets=reshaped_new_y,
        )

    # Expected model
    x_total = torch.vstack([train_x, new_x])  # (n+1,d=1)
    y_total = torch.vstack([train_y, new_y])  # (n+1,n_targets)
    yvar_total = torch.vstack([train_yvar, new_yvar])  # (n+1,n_targets)
    model_expected = SingleTaskGP(x_total, y_total, yvar_total if fixed_noise else None)

    # Check the internal data is the same
    check_same_model_data(model_actual=model_conditioned, model_expected=model_expected)
    # check that they both make the same preditiontion
    x = torch.tensor([[0.4], [0.7]])
    expected_posterior = model_expected.posterior(x)
    conditioned_posterior = model_conditioned.posterior(x)
    check_posterior(conditioned_posterior, expected_posterior)


@pytest.mark.external
# This warning is due to train_y not being standardised. Standarisation can improve quality of fit.
# This is not imporant for this test.
@pytest.mark.filterwarnings(
    "ignore : Input data is not standardized : botorch.exceptions.InputDataWarning : botorch.models"
)
@pytest.mark.xfail(
    reason="Supplying noise in to update homoskedatic has side effects",
)
def test_get_fantasy_model_silent_failure():
    """Demonstate the silent failure when providing noise to a GP that doesn't use it.

    The posteriors produces are not the same. This is an edge case of `test_get_fantasy_model`.
    """
    fixed_noise = False
    n_targets = 1

    train_x = get_train_x()
    n = train_x.shape[0]
    train_y = torch.sin(train_x).expand(n, n_targets)
    train_yvar = torch.ones_like(train_y) * 0.1

    model_original = SingleTaskGP(train_x, train_y, train_yvar if fixed_noise else None)

    # run the model once to avoid runtime error detailing this requirement
    _ = model_original.posterior(torch.tensor([[0.5]]))

    # Generate new training data to use
    new_x = torch.tensor([[0.5]])  # (n,d)
    new_y = torch.tensor([[0.1]]).expand(1, n_targets)  # (n, m)
    new_yvar = torch.ones_like(new_y) * 0.1

    # reshape the new_y and new_yvar to match the format of the underling model. This is required for this method
    # gpytorch docs say: (f,*b, m):
    # In botorch is is written as (*b,n') - which is not super clear because there is no m (target)
    # In reality this just need to match the dimensions of model.train_targets.
    # (n_targets,n) when n_targets> 1
    # (n,) when n_targets==1
    reshaped_new_y = new_y.reshape(-1) if n_targets == 1 else new_y.T
    reshaped_new_yvar = torch.ones_like(reshaped_new_y) * 0.1

    # Shouldn't have noise, but provide it anyway
    model_conditioned = model_original.get_fantasy_model(
        # gpytorch docs say: (f,*b, m, d). In botorch is is written as (*b,n',d)
        inputs=new_x,
        targets=reshaped_new_y,
        # This is Kwarg that gets passed through.  must match new_y
        noise=reshaped_new_yvar,
    )

    # Expected model
    x_total = torch.vstack([train_x, new_x])  # (n+1,d=1)
    y_total = torch.vstack([train_y, new_y])  # (n+1,n_targets)
    yvar_total = torch.vstack([train_yvar, new_yvar])  # (n+1,n_targets)
    model_expected = SingleTaskGP(x_total, y_total, yvar_total if fixed_noise else None)

    # Check the internal data is the same
    check_same_model_data(model_actual=model_conditioned, model_expected=model_expected)
    # check that they both make the same preditiontion
    x = torch.tensor([[0.4], [0.7]])
    expected_posterior = model_expected.posterior(x)
    conditioned_posterior = model_conditioned.posterior(x)
    check_posterior(conditioned_posterior, expected_posterior)



"""This builds of the utils provided in ax.modelbridge.modelbridge_utils."""

import numpy as np
from ax.core.observation import Observation
from ax.modelbridge import modelbridge_utils
from numpy.typing import NDArray


def observations_to_arrays(
    param_names: list[str], outcomes: list[str], observations: list[Observation]
) -> tuple[NDArray[np.float64], NDArray[np.float64], NDArray[np.float64]]:
    """Helper method to convert Ax results (Observation) to Numpy arrays of a standard format.

    This is useful e.g. when converting the output of `ax.model_bridge.base.ModelBridge.get_training_data` into "x" and
    "y" and "covariance" arrays.

    Args:
        param_names: List of the parameter names to report, and in what order.
        outcomes: list of the metric/target names to report, in this order.
            TODO(sw): work out if the order ever changes. If not, we should remove this.
        observations: The list of observations to convert.

    Returns:
        - features: An (n_observation, n_features) array of features.
        - f: An (n_observation, m_metrics) array of prediction means.
        - cov: An (n_observation, m_metrics, m_metrics) array of covariance between metric per x point.

    Example:
        >>> import numpy as np
        >>> from ax.core.observation import Observation, ObservationData, ObservationFeatures
        >>> X = np.array([[1, 2], [3, 4], [5, 6]])
        >>> Y = np.array([[10], [20], [30]])
        >>> of = [ObservationFeatures(parameters={"x1": x[0], "x2": x[1]}) for x in X]
        >>> od = [ObservationData(metric_names=["y"], means=y, covariance=np.eye(1)) for y in Y]
        >>> observations = [Observation(features=f, data=d) for f, d in zip(of, od)]
        >>> features, f, cov = observations_to_arrays(param_names=["x1", "x2"], outcomes=["y"], observations=observations)
        >>> assert np.array_equal(features, X)
        >>> assert np.array_equal(f, Y)
        >>> # There is only one output per x point. This will have perfect covariance with itself.
        >>> assert np.array_equal(cov, np.ones([3, 1, 1]))
    """  # noqa: E501
    features_array = modelbridge_utils.observation_features_to_array(param_names, [o.features for o in observations])
    f, cov = modelbridge_utils.observation_data_to_array(outcomes, [o.data for o in observations])
    return features_array, f, cov



"""Helper for gradient assessment.

NOTE: currently just used as a helper in tests. Could move their if we do not consider it useful to users
"""

import matplotlib.pyplot as plt
import torch
from matplotlib.figure import Figure


def is_smooth_1d(
    x: torch.Tensor,
    y: torch.Tensor,
    d1_threshold: float = 3.0,
    d2_threshold: float = 150,
    *,
    plot: bool = False,
    test: bool = True,
) -> None | Figure:
    """Helper to warn if 1d function is likely not smooth (twice differntiable).

    Args:
        x: (n,) points representing the x (input) values of a function
        y: (n,) points representing the y (output) values of a function
        d1_threshold: Maximum step size allowed in 1st derivative function to be considered smooth.
        d2_threshold: Maximum step size allowed in 1st derivative function to be considered smooth.
        plot: If true, will plot the first and second derivative functions.
        test: is assert statments should be run

    Details
        Smoothness: (defined up to K, the Kth derivative you can take that give a contious funciton over domain)
            - C_0: set of continous functions
            - C_1 (once differentiable):
                - C_0 can be differentiated
                - AND the resulting function is continuous (No steps or holes)
            - ...

        We want C_2 (twice differentiable) for optimisation. As such here we test.
            - if f'(x) is continous.
                - Check if there are not step = not big changes between points.
                - Knowing the appropriate step threshold is hard, the gradient is easier to think about.
                    - Easy way to do this is check if max slope is not exceeded (derivate 2nd)
            - f''(x) is continous.
                - Check if there are not step = not big changes between points
                - Easy way to do this is check if max slope is not exceeded (derivate 2nd)

    NOTE: The thresholds have been set heuristically. It is recomended to plot your function and establish smoothness,
    the use this for regression testing once suitable values have been determined.

    NOTE: The smaller the step size, the better the estimate of gradient (large and small). See y = sim(50 * x)` with
    torch.linspace(0,1,100)` and `torch.linspace(0,1,1000)`
    """
    step_size = torch.diff(x)

    first_derivative = torch.diff(y) / step_size

    if plot:
        fig, axes = plt.subplots(1, 2, figsize=(10, 4))
        _ = axes[0].set_ylabel("1st derivative")
        _ = axes[0].scatter(x[:-1], first_derivative, c="blue", s=2, label="1st derivative")
        _ = axes[0].legend()
        _ = axes[0].set_title("Derivative")

        _ = axes[1].set_title("Derivative Step change")
        _ = axes[1].set_ylabel("1st derivative")
        _ = axes[1].scatter(x[:-2], torch.diff(first_derivative), c="blue", s=2, label="1st derivative")

    d1 = torch.diff(first_derivative).abs() < d1_threshold

    if test:
        msg = (
            f"First derivate is not smooth at indexes {torch.nonzero(~d1).flatten()}.\n These indexes represent the the"
            " 0 based index of 'point_l' in the group of points 'point_l -- point_c -- point_r' in which"
            " the gradient has a step change."
        )
        assert d1.all(), msg

    # each time diff is called, the last element has no next value to calc the diff against
    # as a result that item is dropped
    second_derivative = torch.diff(first_derivative) / step_size[:-1]

    if plot:
        ax1_twin = axes[0].twinx()
        _ = ax1_twin.scatter(x[:-2], second_derivative, c="orange", s=2, label="2nd derivative")
        _ = ax1_twin.set_ylabel("2nd derivative")
        _ = ax1_twin.legend()

        ax2_twin = axes[1].twinx()
        _ = ax2_twin.set_ylabel("2nd derivative")
        _ = ax2_twin.scatter(x[:-3], torch.diff(second_derivative), c="orange", s=2, label="2nd derivative")
        ax2_twin.legend()
        plt.tight_layout()

    d2 = torch.diff(second_derivative).abs() < d2_threshold

    if test:
        assert d2.all(), f"Second derivate is not smooth at indexs {torch.nonzero(~d2).flatten()}"

    return fig if plot else None



"""Fixtures required for this subdirectory."""
# sourcery skip: dont-import-test-modules

import pytest

from tests.sampling import helpers


@pytest.fixture(
    params=[
        helpers.posterior_n1_t1(),
        helpers.posterior_n2_t1(),
        helpers.posterior_n1_t2(),
        helpers.posterior_n3_t2(),
        helpers.posterior_b2_n3_t2(),
    ],
    # provides naming which makes it easi
    ids=[
        "posterior_n1_t1",
        "posterior_n2_t1",
        "posterior_n1_t2",
        "posterior_n3_t2",
        "posterior_b2_n3_t2",
    ],
)
def posterior(request: pytest.FixtureRequest):
    """Provides a generic set of posteriors to be tested against.

    This is useful when functionality should be invariant to specific type of posterior.
    """
    return request.param



"""Helpers built ontop of the distributions module.

Todo:
    These are seperate from `utils` becuase of circular import caused by `ApproximateMixture`. Potentially there is a
    more elegant solution.
"""

import torch
from torch.distributions import Distribution, MixtureSameFamily

from axtreme.distributions.mixture import ApproximateMixture
from axtreme.distributions.utils import dist_dtype


def dist_cdf_resolution(dist: Distribution) -> float:
    """To what resolution (number of decimal places) is cdf output accurate to.

    This is effected by:

        - The numeric precision of the datatype q is stored with (e.g float32)
        - The internal calculation and numerical error incurrred by them.

    Return:
        10**precision, where precision is the decimal position. E.g if accurate to 3 decimal places, return 0.001

    Note:
        - We are interested in the `cdf(x) -> q` accuracy, as this is used heavily.
        - `icdf(q) -> x` does not have the same resolution as detailed here. The difference in results increases with x.
        - This number/method is determined emprically. Tests show it is a good bound for float16 and float32.
            Assumed to hold for float64
    """
    if isinstance(dist, ApproximateMixture):
        return approx_mixture_cdf_resolution(dist)
    if isinstance(dist, MixtureSameFamily):
        return mixture_cdf_resolution(dist)

    dtype = dist_dtype(dist)
    # NOTE: The numeric precision to which q can be represented also appears to bound the cdf output.
    # This has been determined emperically. See unit tests for details.
    # NOTE: this error is larger than just the rounding error we would expect for q (eps/2)
    return torch.finfo(dtype).resolution


def mixture_cdf_resolution(dist: MixtureSameFamily) -> float:
    """To what resolution (number of decimal places) is cdf output accurate to.

    This compared to `dist_cdf_resolution` this is also impacted by `weights*components`.

    This is effected by:

        - The numeric precision of the datatype q is stored with (e.g float32)
        - The internal calculation and numerical error incurred by them.
        - Combination of `weight*components`

    Return:
        10**precision, where precision is the decimal position. E.g if accurate to 3 decimal places, return 0.001

    Note:
        - We are interested in the `cdf(x) -> q` accuracy, as this is used heavily.
        - `icdf(q) -> x` does not have the same resolution as this. The difference in results increases with x.
        - This number/method is determined emprically. Tests show it is a good bound for float16 and float32.
          Assumed to hold for float64
    """
    dtype = dist_dtype(dist)
    # NOTE: The numeric precision to which q can be represented also appears to bound the cdf output.
    # This has been determined emperically. See unit tests for details. As per the unit test, we add a safety factor of
    # 10, as we are not certain the unittest represents worst case behaviour.
    # NOTE: this error is larger than just the rounding error we would expect for q (eps/2)
    return torch.finfo(dtype).resolution * 10


def approx_mixture_cdf_resolution(dist: ApproximateMixture) -> float:
    """To what resolution (number of decimal places) is cdf output accurate to.

    This is identical to `mixture_cdf_resolution` except for the conservatism introduced by the approximation.
    See `ApproximateMixture` "Impact of Approximation:" for details
    """
    dtype = dist_dtype(dist)
    resolution = mixture_cdf_resolution(dist)
    # As per ApproximateMixture the approximation error is bounded by finfo.eps
    return max(resolution, torch.finfo(dtype).eps)



import numpy as np
import pytest
import torch
from torch.utils.data import DataLoader, SequentialSampler

from axtreme.data import MinimalDataset, MultiBatchSampler

# what do we want to check on the multibatch sampler?

# Bad batch shape give an error
# That the content of the batchs changes with different sample size


# Usecase tests: (this well go in a demofile)
# Have a lot of data, use a subset to make something and iterate though
# if you intend to combine it show that should use the other one
# What do do if you want it iterate through a whole dataset, and want to do it fast.
# (recommended just make it fit the batch shape for now)


@pytest.mark.parametrize(
    "batch_shape",
    [
        torch.Size([12]),
        torch.Size([6]),
        torch.Size([4, 3]),
        torch.Size([2, 3]),
        torch.Size([2, 3, 2]),
    ],
)
def test_multibatchsampler_full_batchs(batch_shape: torch.Size):
    # sourcery skip: no-loop-in-tests
    # Typically this will be a Sampler object, but an iterator is perfectly valid, and is simpler to is used here
    samples = list(range(12))
    batch_sampler = MultiBatchSampler(samples, batch_shape=batch_shape)

    for batch in batch_sampler:
        assert torch.tensor(batch).shape == batch_shape


@pytest.mark.parametrize(
    "batch_shape, partial_batch_dim",
    [
        # 1d check
        (torch.Size([5]), -1),  # partial batch of size torch.Size([2])
        # 2d check
        (torch.Size([2, 5]), -1),  # partial batch of size torch.Size([2,1])
        (torch.Size([5, 2]), 0),  # partial batch of size torch.Size([1,2])
        # 3d check
        (torch.Size([2, 2, 2]), -1),  # partial batch of size torch.Size([2,2,1])
        (torch.Size([2, 2, 2]), 1),  # partial batch of size torch.Size([2,1,2])
    ],
)
def test_multibatchsampler_partial_batchs(batch_shape: torch.Size, partial_batch_dim: int):
    # sourcery skip: no-loop-in-tests
    # Typically this will be a Sampler object, but an iterator is perfectly valid, and is simpler to is used here
    samples = list(range(12))
    batch_sampler = MultiBatchSampler(samples, batch_shape=batch_shape, partial_batch_dim=partial_batch_dim)

    # Check the standard batches
    batches = list(batch_sampler)
    for batch in batches[:-1]:
        # print this if want more viual debugging
        assert torch.tensor(batch).shape == batch_shape

    # check the the final batch has the same dimensionf for all but the partial_batch_dim
    partial_batch = batches[-1]
    partial_batch_shape = torch.tensor(partial_batch).shape
    # remove the partial index from each and check that shape matches otherwise
    batch_shape_list = list(batch_shape)
    partial_batch_shape_list = list(partial_batch_shape)

    del batch_shape_list[partial_batch_dim]
    del partial_batch_shape_list[partial_batch_dim]

    assert batch_shape_list == partial_batch_shape_list


@pytest.mark.parametrize(
    "batch_shape, data_length,  partial_batch_dim",
    [
        # 1d check
        (torch.Size([5]), 12, None),
        # # 2d check
        (torch.Size([5, 2]), 12, -1),  # partial batch of size torch.Size([2,1])
        (torch.Size([2, 5]), 12, 0),  # partial batch of size torch.Size([1,2])
    ],
)
def test_multibatchsampler_incompatible_partial_batchs(
    batch_shape: torch.Size, data_length: int, partial_batch_dim: int
):
    """We know ahead of time when the partial batch will fail, because we only allow one dime to be parial.

    e.g if trying to fit [1,2,3,4,5] into shape
        [[.,.],
         [.,.],
         [.,.]]
    """
    samples = list(range(data_length))

    with pytest.raises(ValueError, match="..* does not fit into batch shape,*"):
        _ = MultiBatchSampler(samples, batch_shape=batch_shape, partial_batch_dim=partial_batch_dim)


def test_multibatchsampler_no_batch_invariant():
    """Samples in batchs are not invariant to the partial_dim (which is typically the on batching along).

    The follow show how the data along the partial_dimension (rows) changes depending on the size.
    This will cause variance between runs with different batch size if you later operate along this dim (e.g rows).

    See BatchSampler2dInvariant for details.
    """
    samples = list(range(8))
    # What ould happen if full size batching was used
    batch_sampler2 = MultiBatchSampler(samples, batch_shape=torch.Size([2, 4]))
    samples2 = torch.concat([torch.tensor(b) for b in batch_sampler2], dim=-1)

    expected_result2 = torch.tensor([[0, 1, 2, 3], [4, 5, 6, 7]])

    torch.testing.assert_close(samples2, expected_result2)

    # What will happen when batch dimension is changed

    batch_sampler1 = MultiBatchSampler(samples, batch_shape=torch.Size([2, 2]))
    samples1 = torch.concat([torch.tensor(b) for b in batch_sampler1], dim=-1)

    expected_result1 = torch.tensor([[0, 1, 4, 5], [2, 3, 6, 7]])

    torch.testing.assert_close(samples1, expected_result1)


### Integration test
def test_multibatchsampler_integration_compatible_with_multiworker():
    """Integration test to check that works with the standard components *e.g Sampler and Dataloader.

    NOTE: any call starts iterating through a dataloader object, need to be inside __name__=="__main__"
    (more specifically, must not run on import). This stops the subprocess for running it again (creating a whole new
    set of worker).
    IF you set num_workers = 0 then don't need to wory about this as work is just done in current process.
    """
    # sourcery skip: no-loop-in-tests, remove-empty-nested-block
    # make a fake dataseet where each feature has 2 dims
    data = np.arange(0, 200).reshape(-1, 2)
    ds = MinimalDataset(data)
    sampler = SequentialSampler(ds)

    batch_sampler = MultiBatchSampler(sampler, batch_shape=torch.Size([4, 3]))
    dl = DataLoader(ds, batch_sampler=batch_sampler, num_workers=2)

    for _ in dl:
        pass



import numpy as np
from numpy.typing import NDArray

from axtreme.simulator import utils


def test_is_valid_simulator():
    def foo(
        x: np.ndarray[tuple[int, int], np.dtype[np.float64]], n_simulations_per_point: int = 1
    ) -> np.ndarray[tuple[int, int, int], np.dtype[np.float64]]:
        raise NotImplementedError

    assert utils.is_valid_simulator(foo, verbose=True)


def test_is_valid_simulator_incorrect_params():
    def foo(
        x: np.ndarray[tuple[int, int], np.dtype[np.float64]],
    ) -> np.ndarray[tuple[int, int, int], np.dtype[np.float64]]:
        raise NotImplementedError

    assert utils.is_valid_simulator(foo) is False


def test_is_valid_simulator_incorrect_types():
    def foo(
        x: np.ndarray[tuple[int, int], np.dtype[np.float64]], n_simulations_per_point: str
    ) -> np.ndarray[tuple[int, int, int], np.dtype[np.float64]]:
        raise NotImplementedError

    assert utils.is_valid_simulator(foo) is False


def test_is_valid_simulator_incorrect_missing_default():
    def foo(
        x: np.ndarray[tuple[int, int], np.dtype[np.float64]], n_simulations_per_point: int
    ) -> np.ndarray[tuple[int, int, int], np.dtype[np.float64]]:
        raise NotImplementedError

    assert utils.is_valid_simulator(foo) is False


def test_is_valid_simulator_incorrect_return():
    def foo(x: np.ndarray[tuple[int, int], np.dtype[np.float64]], n_simulations_per_point: int = 1) -> bool:
        raise NotImplementedError

    assert utils.is_valid_simulator(foo) is False


def test_is_valid_simulator_allow_custom_defaults():
    def foo(
        x: np.ndarray[tuple[int, int], np.dtype[np.float64]], n_simulations_per_point: int = 10
    ) -> np.ndarray[tuple[int, int, int], np.dtype[np.float64]]:
        raise NotImplementedError

    assert utils.is_valid_simulator(foo)


def test_simulator_from_func_adds_n_simulations_per_point_correctly():
    def dummy_function(x: NDArray[np.float64]) -> NDArray[np.float64]:
        return x + 1

    simulator = utils.simulator_from_func(dummy_function)

    x = np.array([1, 10]).reshape(-1, 1)
    expected_output = np.array([2, 2, 11, 11]).reshape(2, -1, 1)

    output = simulator(x, n_simulations_per_point=2)

    assert np.allclose(output, expected_output)



"""Helpers and documentation for understanding the impacts of numerical precision."""

import torch


def maximal_representation_error_between_0_1(dtype: torch.dtype) -> float:
    """Maximal amount a float can be wrong by, for numbers within the range [0,1).

    NOTE: This is calculated for the range [.5,1). Errors will be smaller for ranges < .5.

    Args:
        dtype: The dtype to calculate the maximal representation error for.

    Returns:
        The maximal representation error for numbers between 0 and 1.

    Details:
        Mantisa (significant bits) with `x` explicity bits internally has 1 implict bit added. Effectively this become
        1.[x bits] (in binary).

            - e.g float32 has 24 bits of precision but 23 bits of mantisa

        Epsilon, the smallest **representable** number where 1+eps != 1 can be calculated as `2^-x`

            - e.g float32: 2^-23 = 1.19e-7

        The largest representable number less than 1 is calculated as: `1 - 2^-(x + 1)`:

            - e.g float32: 1 - 2^-24 = 0.999999940395355225

        As such the smallest representable number between [.5,1) is:  `2^-(x + 1)`

            - e.g float32:  2^-24 = 5.96e-08

        Value smaller than the smallest representable number will be rounded to the nearest representable number. As
        such the largest representation error for number between [.5,1) is `2^-(x + 1)/2` or `2^-(x_2)`

            - e.g float32:  2^-(23+2) = 2.9802322387695312e-08

        Resoltuion and Precisions:
        - TODO(sw): fill out exactly how these work. These number are ignored for now.

        Additional information:

            - https://en.wikipedia.org/wiki/Single-precision_floating-point_format
    """
    mantisa_bits = {
        torch.float16: 10,
        torch.float32: 23,
        torch.float64: 52,
    }
    return 2 ** -(mantisa_bits[dtype] + 1 + 1)



"""Mixture model variants."""

import torch
from torch.distributions import Categorical, Distribution, MixtureSameFamily

from axtreme.distributions.utils import dist_dtype


class ApproximateMixture(MixtureSameFamily):
    """Mixture distributions where extreme caclulations are approximated.

    Some distribution only support a limited range of quantiles (e.g :math:`[.01, 99]`) due to numerical issues (see
    "Details"). When calculation such as :math:`q=cdf(x)` or :math:`x=icdf(q)` fail outside this range the function then
    error. ``ApproximateMixture`` allows all x values ,and approximates the results for x values outside the supported
    range (see "Details" for approximation method).

    Details:

        Distribution Quantiles Bounds:

            Some distributions (e.g `TransformedDistribution`) have bounds on the quantiles they can support.
            Calculations such as :math:`q=cdf(x)` or :math:`x=icdf(q)` will fail when q is outside of these bounds.
            Bounds exist because for sum distributions :math:`icdf(1)=inf` or :math:`icdf(0)=-inf`. Eventually there
            comes a point for q value close to 0 or 1 where they lack the numerical precision to capture very small
            changes in q, and the very large values of x.

        Approximation principles:

            It is assumed we want to conservatively estimate (:math:`1-cdf(x)`), the chance the a value will exceed some
            level :math:`x`. For example, x could represent the strength a structure is designed to withstand, and
            (:math:`1-cdf(x)`) represents the chance of experiencing a force that will break the structure (e.g the
            risk). It is better to overestimate the risk (conservative), rather than underestimate it. In other words,
            if the :math:`cdf_est(x)` over estimates the :math:`cdf_true(x)`, then the true risk of exceeding x is. e.g:

            TLDR:

                - :math:`cdf_est(x) < cdf_true(x)`: produces conservative design (okay)
                - :math:`cdf_est(x) > cdf_true(x)`: BAD

            Worked example:

                - :math:`cdf_est(x) = .5` and :math:`cdf_true(x) = .4`
                - If structure is designed to be :math:`x` strong, then estimated number of failure is .5, true number
                  of failures is .6.
                - Have underestimated the risk and designed a structure more likely to fail than we expect.

        Approximate results.

            ApproximateMixture provides exact results within the quantile bounds ``[finfo.eps, 1 - finfo.eps]`` (where
            ``finfo = torch.finfo(component_distribution.dtype)``) For details regarding why this range is selected see
            ``_lower_bound_x`` and ``_upper_bound_x``. Values smaller than ``finfo.eps`` are approximated
            according to ``_lower_bound_x``. Values larger than ``1 - finfo.eps`` are approximated according to
            ``_upper_bound_x``. Note the range ``[finfo.eps, 1 - finfo.eps]`` is still used even if the component
            distribution supports a greater range.

        Impact of Approximation:

            As a result of the approximation, the cdf method can underestimate the analytical cdf value by up to
            ``finfo.eps``. This values come from the following calculation, Similar behaviour occurs at the lower bound.

                - Consider one of the underlying distributions in the marginal: once x hits the upper bound

                    - it will give q = 1.0-finfo.eps
                    - in reality it could be as high as 1.0
                    - so the q give is finfo.eps too small (at worst)

                - The marginal cdf is calcualted from the underlying distributions:

                    - e.g :math:`q_marginal = w_1 * q_1 + ... + w_n * q_n`
                    - :math:`sum(w_i) = 1`

                - In the worst case:

                    - Underling distribution: q give is finfo.eps too small (at worst)
                    - weight = 1

    Note:
        It is worth ensuring that the ApproximateMixture distribution has suitable numeric precision for its intended
        use. See ``axtreme.distributions.utils.mixture_dtype`` for more details.
    """

    def __init__(
        self,
        mixture_distribution: Categorical,
        component_distribution: Distribution,
        *,
        validate_args: bool | None = None,
        check_dtype: bool = True,
    ) -> None:
        """Initialise the ApproximateMixture.

        Args:
            mixture_distribution: ``torch.distributions.Categorical``-like instance. Manages the probability of
              selecting component. The number of categories must match the rightmost batch dimension of the
              ``component_distribution``. Must have either scalar ``batch_shape`` or ``batch_shape`` matching
              ``component_distribution.batch_shape[:-1]``
            component_distribution: ``torch.distributions.Distribution``-likeinstance. Right-most batch dimension
              indexes component.
            validate_args: Runs the default validation defined by torch
            check_dtype: Check datatype of distributions match and inpput are at least as precise as the distribution.
        """
        super().__init__(mixture_distribution, component_distribution, validate_args=validate_args)
        self.check_dtype = check_dtype

        component_dtype = dist_dtype(component_distribution)
        mixture_dtype = dist_dtype(mixture_distribution)
        if check_dtype and (component_dtype != mixture_dtype):
            msg = (
                "Component and mixture distributions should be of the same dtype, otherwise results could have the"
                f" resolution of the lowest dtype. Got dtypes {component_dtype} and {mixture_dtype} respectively"
            )
            raise TypeError(msg)
        self.dtype = component_dtype

        finfo = torch.finfo(component_dtype)
        # NOTE: Even though internally TranformedDistribution sets its lowerbound with finfo.tiny, it appears to have
        # issues below finfo.eps. See `docs\source\marginal_cdf_extrapolation.md` "Distribution lower bound issue"
        self.quantile_bounds = torch.tensor([finfo.eps, 1 - finfo.eps], dtype=component_dtype)
        # NOTE: bounds are stored here so we don't need to be recomputed every `.pdf()` or `.cdf()` call.
        # Can put in the call if we want to save memory.
        self.x_bound_lower = ApproximateMixture._lower_bound_x(component_distribution)
        self.x_bound_upper = ApproximateMixture._upper_bound_x(component_distribution)

    @staticmethod
    def _lower_bound_x(dist: Distribution) -> torch.Tensor:
        """Returns the x lowerbound for each of the component distributions.

        If :math:`cdf(x)` recieves values that are too small, it throws an error. We instead find a suitable lower
        bound and later clip the x values to this range. The lower bound should:

            - Not throw an error when it is used in :math:`cdf(lower_bound).`
            - As per "Approximation principles" it should not result in overestimate of the analytical cdf.

        Args:
            dist: The component distribution.

        Return:
            Tensor of shape (dist.batch_shape).

        Details:
            TransformedDistribution are officially bounded between ``(finfo.tiny, 1 - finfo.eps)``. The CDF does not
            match the mathematical CDF across this whole region (due to numeric errors) - Specifically the region
            (finfo.tiny, finfo.eps).

            in the region (finfo.tiny, finfo.eps):

                - :math:`cdf(x)`: will output 0 for most of the region
                - :math:`cdf(x)`: `x` corresponding q = finfo.tiny will produces error about being outside of the
                  allowed range,
                  due to numerical issues. :math:`x` in the region will produce the same error.

            Solution: Pick an x value corresponsing to the middle q range (finfo.tiny, finfo.eps). This should produce
            cdf(x)=0, which underestimates the true cdf value by up to `finfo.eps`. As per "Approximation principles"
            this is acceptable.
        """
        dtype = dist_dtype(dist)
        finfo = torch.finfo(dtype)
        tiny = torch.tensor(finfo.tiny, dtype=dtype)
        eps = torch.tensor(finfo.eps, dtype=dtype)

        # See unit test visualisation for whne a q between tiny and eps can't be used directly
        return (dist.icdf(tiny) + dist.icdf(eps)) / 2

    @staticmethod
    def _upper_bound_x(dist: Distribution) -> torch.Tensor:
        """Returns the x upperbound for each of the component distributions.

        If :math:`cdf(x)` recieves values that are too large, it throws an error. We instead find a suitable upper bound
        and later clip the x values to this range. The upper bound should:

            - Not throw an error when it is used in :math:`cdf(upperbound)`.
            - As per "Approximation principles" it should not result in overestimates of the analytical cdf.

        Args:
            dist: The component distribution.

        Return:
            Tensor of shape (``dist.batch_shape``).

        Details:
            TransformedDistribution are officially bounded btween ``(finfo.tiny, 1 - finfo.eps)``. There do not appear
            to be issues using :math:`cdf(x)` for x values corresponding to slighly larger than ``1-finfo.eps``. These
            produce values up to 1.

            NOTE: IF we can confirm distribution behaviour is safe outside of the offical bounds specified on the
            distribution (``1-finfo.eps``) this value could be increased. Currently we underestimates any value outside
            ``1-finfo.eps`` as we are unsure of the behaviour within the range ``[1-finfo.eps, 1]``.As per
            "Approximation principles" this is acceptable.
        """
        dtype = dist_dtype(dist)
        finfo = torch.finfo(dtype)
        eps = torch.tensor(finfo.eps, dtype=dtype)

        return dist.icdf(1 - eps)

    def cdf(self, x: torch.Tensor) -> torch.Tensor:
        """Return the CDF.

        Identical to MixtureSameFamily implementation except for clamping.

        Args:
            x: Values to calcuate the CDF for. Must be broadcastable with the ``ApproximateMixture.batch_shape``. E.g

                - ``self.component_distribution.batch_shape = (2,5)`` (last dimension is the components that are
                  combined to make a single Mixture distribution)
                - ``self.batch_shape = (2,)``
                - ``x.shape = (10)`` This will fail as it is not broadcastable
                - ``x.shape = (10,1)`` This will pass as it is broadcastable

        Returns:
            Tensor of shape (x.shape[:-1], self.batch_shape)
        """
        if self.check_dtype and torch.finfo(x.dtype).resolution > torch.finfo(self.dtype).resolution:
            msg = (
                f"Input type {x.dtype} is less precise than dtype of the distribution {self.dtype}."
                " This can lead to loss of precision."
            )
            raise TypeError(msg)

        # Expected x already broadcasts to `self.batch_dimension`.
        # Padding is added so x broadcasts to each of the component distributions that make up the mixture
        x = self._pad(x)
        clipped_x = torch.clamp(x, self.x_bound_lower, self.x_bound_upper)
        cdf_x = self.component_distribution.cdf(clipped_x)
        mix_prob = self.mixture_distribution.probs

        # This is the approach found in torch. Can introduce small numerical errors, but we assume these are neglible.
        return torch.sum(cdf_x * mix_prob, dim=-1)

    def log_prob(self, x: torch.Tensor) -> torch.Tensor:
        """Calculate the log prob (e.g log(pdf)).

        Args:
            x: Values to calcuate the log prob for. Must be broadcastable with the ``ApproximateMixture.batch_shape``.
              E.g

                - ``self.component_distribution.batch_shape = (2,5)`` (last dimension is the components that are
                  combined to make a single Mixture distribution)
                - ``self.batch_shape = (2,)``
                - ``x.shape = (10)`` This will fail as it is not broadcastable
                - ``x.shape = (10,1)`` This will pass as it is broadcastable

        Returns:
            Tensor of shape (x.shape[:-1], self.batch_shape)
        """
        # Dev notes:
        # NOTE: MixtureSameFamily.log_prob makes use of `_validate_sample(x)`
        # There are challenges with the shape here because:
        # - Background:
        #     - Mixture has shape (*b,c) where:
        #         - *b: is batch dim, each of them representing a different mixture dist
        #         - c: is the component dists that get summed together to make a mixture
        #     - X has shape (*s,*b) where:
        #         - *s: batches of data to evaluate with each mixture dist
        #         - *b: matches the batch (number of models) of the mixture (or can be broadcast)
        #     e.g: to feed 10 input to 2 different mixtures the shapes would be
        #         - Mixture: (b  = 2, c = n)
        #         - x = (s = 10, b = 1)  -- the batch dimesnions can boudcast
        # - Our context:
        #     - Each x item need to be cliped to the range of the component distribution `c`
        #     - x_clipped = (*s,*b,c)
        #     - This shape is not compatible with _validate_sample(clipped_x)
        #         - this expects shape (*b)
        #             - First it first does check to make sure we get a specific shape (irrelevant to us)
        #             - then it broadcast this to (*s,*b,c) and checks if within supported range

        # - required decision:
        #     - Decide if we need to use _validate_sample, and if so, which parts.
        raise NotImplementedError


def icdf_value_bounds(dist: MixtureSameFamily, q: torch.Tensor) -> torch.Tensor:
    r"""Returns bounds in which the value for quantile q is gaurenteed to be found.

    Args:
        dist: ``(*batch_shape,)`` mixture distribution producing events of event_shape samples.
        q: quantile to find the inverse cdf of. Must be boardcastable up to (``*batch_shape``,). Must not have more
           dimensions than ``*batch_shape`` (only 1 q can be passed to each of the distributions in the batch.)

    Returns:
        tensor of shape (2,*batch_shape), there the first index represents the lower
        bounds, and the second the upper bounds.

    Details:
    Mixture distribution calculate the CDF as follows:

        :math:`q = w_i * CDF_1(y) + w_2 * CDF_2(y) + ... + w_n * CDF_n(y)`
        Which can be written as: :math:`q = w_i * q_1 + w_2 * q_2 + ... + w_n * q_n`

        where :math:`0 <= w_i <= 1` and :math:`\\sum{w_i} = 1`

    An effective way to bound the x values the can produce y is:

        - take the ``icdf(q)`` for each distribution. Now have X_n values.
        - ``lower_bound = min(X_n)``: at this point the first component distribution has become big enough to produce q.
          As the weights are between [0,1] no point prior would be able to procude q as no q_i was large enough.s
        - ``upper_bound = max(X_n)``: at this point the last component distribution has become big enough to produce q.
          As the weights sum to one, q must be produced by this point.
    """
    # expand can not be done to smaller shapes, so this safely handled the case when q.dim() > len(dist.batch_shape)
    q_expanded = q.expand(dist.batch_shape)

    q_expanded = q_expanded.unsqueeze(-1)  # add a dimension for the underlying component distributions.
    values = dist.component_distribution.icdf(q_expanded)
    lower_bound = values.min(dim=-1).values
    upper_bound = values.max(dim=-1).values

    return torch.stack([lower_bound, upper_bound])



"""Helper for finding inverse cdf for function that do not have an icdf method."""

from itertools import product

import torch
from scipy.optimize import RootResults, root_scalar
from torch.distributions import Distribution

from axtreme.distributions.helpers import dist_cdf_resolution
from axtreme.distributions.utils import dist_dtype, index_batch_dist

# pyright: reportUnnecessaryTypeIgnoreComment=false


def icdf(
    dist: Distribution,
    q: torch.Tensor,
    max_acceptable_error: float,
    opt_bounds: torch.Tensor,
) -> torch.Tensor:
    """Calculated the inverse CDF for each distribution in a batch.

    This method is useful when the distrbution does not have an icdf method. In this case the CDF can be run repeatidly
    until we find the `x` where `cdf(`x`) == quantile`. We use optimisation to make this search effecient.

    Args:
        dist: (`*batch_shape`,) mixture distribution producing events of event_shape samples
        q: quantile to find the inverse cdf of. Must be boardcastable up to (`*batch_shape`,). Must not have more
           dimensions than `*batch_shape` (only 1 q can be passed to each of the distributions in the batch.)
        opt_bounds:(2,`*batch`). The lower and lower x bounds withing which the q can be found. If unknown, set to
            a very large range. `opt_bounds.shape[1:]` must be broadcastable to `dist.batch_shape`.
        max_acceptable_error: return values must be within `icdf(q +/- max_acceptable_error)`

    Returns:
        Tensor of shape (dist.batch_shape) of the icdf(x) results.
    """
    # Expand insure q.dim() < len(dist.batch_shape)
    q_expanded = q.expand(dist.batch_shape)
    bounds_expanded = opt_bounds.expand(torch.Size([2]) + dist.batch_shape)

    idx_ranges = [range(dim) for dim in dist.batch_shape]
    # List of all the indexes in the batch shape, in the same order as tensor.flatten()
    idx_list = product(*idx_ranges)

    # order of items matches flatten.
    results = []
    for idx in idx_list:
        single_dist = index_batch_dist(dist, idx)

        bounds = bounds_expanded[:, *idx]  # type: ignore  # noqa: PGH003 mypy doesn't recognise :

        result = icdf_1d(
            dist=single_dist,
            quantile=q_expanded[idx].item(),
            max_acceptable_error=max_acceptable_error,
            bounds=tuple(bounds.tolist()),
        )
        results.append(result)

    return torch.tensor(results).reshape(dist.batch_shape)


def icdf_1d(
    dist: Distribution,
    quantile: float,
    max_acceptable_error: float,
    bounds: tuple[float, float],
) -> torch.Tensor:
    """Calculates inverse CDF values for distributions which do not have a icdf function.

    Some distributions do not have an inverse CDF function. In this case the CDF can be run repeatidly until we find the
    `x` where `cdf(`x`) == quantile`. We use optimisation to make this search effecient. This function only supports
    distribution with 1d input and output.

    Args:
        dist: This expects a Distribution like object. The specific methods and attributes required are:

            - methods: `cdf()`, and potentially `log_prob()` depending on the optimiser used.
            - Attributes: None.

        quantile: The quantile for which to find the corresponding x value.
        max_acceptable_error: return values must be within `icdf(q +/- max_acceptable_error)`
        bounds: The bounds to serach for the root in

    Returns:
        A tensor of shape (1,) of the x value corresponding to the quantile.

    Details:
        There are a number of way numerical issues can effect this results:
            - Case 1: Function output (dist.cdf) is not precise enough to support max_acceptable_error.
            - Case 2: Function internals are not precise enough to support max_acceptable_error (e.g they introduce
              numeric error).
            - Case 3: Optimiation input (x), is not accurate enough to support step sizes need to achieve
              max_acceptable_error.
            - Case 4: The optimisation process does not find a suitable result.
            - Case 5: The output datatype truncation is not precise enough, failing max_acceptable_error after
              truncation.
    """
    dtype = dist_dtype(dist)

    # Checks case 1 and 2.
    if dist_cdf_resolution(dist) > max_acceptable_error:
        msg = (
            "The distribution provided does not have suitable resolution (numeric precision) for the"
            f" max_accptable_error specified. {dist_cdf_resolution(dist) = } and {max_acceptable_error = }"
            "increase the max_acceptable_error or use bigger dtype in the distribution."
        )
        raise TypeError(msg)

    # Scipy operates in float64. If scipy inputs are clipped to float32 before the function is run it can cause issues.
    def f(x: float) -> float:
        # TODO(sw 2024-12-17): Does converting to float64 mean the cdf automatically runs in float64? Is the reasonable
        # if users specifically request a float32 is used?.
        x_tensor = torch.tensor([x], dtype=torch.float64)
        return (dist.cdf(x_tensor) - quantile).item()

    # Scipy doesn't have an option to set tolerance in terms of the output, only in terms of x. Output tolerance is the
    # termination criteria we care about, so we need to check this manually.
    for maxiter in [30, 100, 300]:
        # NOTE: better documentation of bisect args can be found scipy.optimize._zeros_
        # NOTE: root_scalar overides the following arguments {disp = False, full_output = True}
        opt_result: RootResults = root_scalar(
            f,
            bracket=bounds,
            method="bisect",
            maxiter=maxiter,
        )

        # NOTE: "output resolution check" gaurentees the distribution has enough accuracy to produce the required output
        # here we use the same dtype as scipy operates in, as we don't want to truncate the input. The intent here is to
        # see if a good enough solution has been found. If there are tunction errors with the chosen data type, that
        # should be raised later.
        x64 = torch.tensor(opt_result.root, dtype=torch.float64)
        q_error_opt = dist.cdf(x64) - quantile

        if q_error_opt.abs() < max_acceptable_error:
            # Termination criteria has been met.
            break

    # Check case 3:
    # If converged, means we have hit the minimum x step size available for that number
    # NOTE: Scipy automatically terminates when the minimum numeric precision is hit for an x value. This scales with
    # the magnitude of x.
    if (q_error_opt.abs() > max_acceptable_error) and opt_result.converged:
        msg = (
            "Optimisation has reached minimum step size in x, but the error is still greater than max_acceptable_error."
            " The datatype used here is python/numpy float - increasing these beyond float 64 is not currently"
            " supported. Increase max_acceptable_error."
        )
        raise RuntimeError(msg)

    # Check case 4:
    # NOTE: Due to case 1 and 2 gauranteeing the output is accurate enough, this is suffecient to check optimiation
    # result.
    if q_error_opt.abs() > max_acceptable_error:
        msg = (
            f"Optimisation result has error  ({q_error_opt})  greater than  {max_acceptable_error=}."
            " Is may be cause by insuffecient optimsation iterations"
        )
        raise RuntimeError(msg)

    # Check case 5:
    x = torch.tensor(opt_result.root, dtype=dtype)
    q_error = dist.cdf(x) - quantile
    if (q_error_opt.abs() < max_acceptable_error) and (q_error.abs() > max_acceptable_error):
        msg = (
            f"Succesful optimisation results fail once trucated to the specifed dtype ({dtype =})."
            " Increase max_acceptable_error or dist.dtype."
        )
        raise TypeError(msg)

    return x



import numpy as np
import torch
from ax import ParameterType, RangeParameter, SearchSpace
from ax.core.observation import Observation, ObservationData, ObservationFeatures
from ax.modelbridge.modelbridge_utils import (
    array_to_observation_data,
    observation_data_to_array,
    observation_features_to_array,
)
from ax.modelbridge.transforms.standardize_y import StandardizeY
from ax.modelbridge.transforms.unit_x import UnitX
from numpy.testing import assert_allclose

from axtreme.utils import transforms


# TODO(sw): Come back and set this up for a multi input and output to ake sure the shapes are right
def test_translate_standardisey():
    ob_datas = array_to_observation_data(
        f=np.array(
            [
                [1, 7],
                [2, 6],
            ]
        ),
        cov=np.array(
            [
                [
                    [0.5, 0.1],
                    [0.1, 0.5],
                ],
                [
                    [1, 0.1],
                    [0.1, 2],
                ],
            ]
        ),
        outcomes=["a", "b"],
    )
    obs = [Observation(features=ObservationFeatures(None), data=obs_data) for obs_data in ob_datas]

    # Standardises the y for each metric
    transform = StandardizeY(observations=obs)

    input_transform, output_transform = transforms.translate_standardisey(transform, col_names=["a", "b"])
    assert input_transform is None
    assert output_transform is not None

    # we care if the transform works, not the internals, so we test the output
    ### Set up the data in model space to be untransformed
    # data for ax
    obs_model_space = Observation(
        features=ObservationFeatures(None),  # Note used, put none to make this obvious
        data=ObservationData(
            metric_names=["a", "b"],
            means=np.array([3, 8]),
            covariance=np.array(
                [
                    [1, 0.1],
                    [0.1, 2],
                ],
            ),
        ),
    )

    # same data for botorch
    y, yvar = observation_data_to_array(outcomes=["a", "b"], observation_data=[obs_model_space.data])
    y = torch.tensor(y)
    yvar = torch.tensor(yvar)

    ### Transform each to the problem space
    # ax
    obs_data_problem_space_expected = transform.untransform_observations([obs_model_space])[0].data

    # Make the equivalent model space value in tensor form
    botorch_output_means, botorch_output_cov = output_transform.untransform(Y=y, Yvar=yvar)
    assert botorch_output_cov is not None

    assert_allclose(obs_data_problem_space_expected.means, botorch_output_means[0].numpy())
    assert_allclose(obs_data_problem_space_expected.covariance, botorch_output_cov[0].numpy())


def test_translate_unitx():
    junk_search_space = SearchSpace(
        parameters=[
            RangeParameter(name="a", parameter_type=ParameterType.FLOAT, lower=0, upper=50),
            RangeParameter(name="b", parameter_type=ParameterType.FLOAT, lower=0, upper=25),
        ]
    )

    transform = UnitX(junk_search_space)

    input_transform, output_transform = transforms.translate_unitx(transform)
    assert input_transform is not None
    assert output_transform is None

    #### Create the feature to be transofrmed
    feature_ax = ObservationFeatures(
        parameters={
            "a": 25,
            "b": 12.5,
        }
    )
    feature_botorch = torch.tensor(observation_features_to_array(["a", "b"], [feature_ax]))

    ### Transform the features
    ax_transformed = transform.transform_observation_features([feature_ax])
    botorch_transformed = input_transform.transform(feature_botorch)

    ### Check values are equivalent
    assert_allclose(observation_features_to_array(["a", "b"], ax_transformed), botorch_transformed.numpy())



"""Responsible for running simulators with a given set of Ax Parameters, and returning estimate with uncertainty."""

from dataclasses import dataclass

import numpy as np
from ax.core import TParameterization
from scipy.stats import rv_continuous

from axtreme.simulator import Simulator
from axtreme.utils import distibution_helpers


# QUESTION: is this the right place for this?
@dataclass
class SimulationPointResults:
    """For a point that has been simulation, stores the mean(s) and covariance(s).

    Args:
        means (NDArray[np.float64]): array of the mean paramter/metric estimates at a X point.
        cov (NDArray[np.float64]|None): covarianc matrix with uncertainty distibution of the metric estimates.
            - This can be None if the error is unknown
        parameter_names: list of names. Gives the index where relevant data is stored

    Design rational:
        - The primary purpose of this is to define the interface between the `Runner` which generate simulation results,
          and `Metric` which reports the required parts of the results for AX to then use.
        - The intent is to :
            - Explicitally define the interface information between the two (rather than use a dict)
            - Prevent `Runner` needing to know about specific structure of metric.
            - Prevent `Metric` needing to know the stucture of Runner
            - Esentially it mean the translatioin logic between these two components is contained in one discrete
              unit/object


    This is generated (in the `Runner`) when the simulation is evaluated for a specific `Trial`.
    It is attached to the `Trail.metadata`, and later read by `Metric.fetch_trial_data`.

    Args:
        - metric_name: The name of the metric that these results are relevant to
        - mean: The mean estimate of the metric for this particular trial
        - sem: Standard Error Measure, as defined `here <https://ax.dev/docs/glossary.html#sem>`_

    Background: For documenation on what can do into this object
        - `Ax tutorial on defining custom metrics <https://ax.dev/tutorials/gpei_hartmann_developer.html#8.-Defining-custom-metrics>`_
        - `Ax definition <https://ax.dev/docs/glossary.html#sem>`_
        - `For the possible things might want to contruct in the end <ax.core.data.Data.REQUIRED_COLUMNS>`_

    Todo:
        Revist if there is a better abstraction for this. Detail in Github issue #31.
    """

    metric_names: list[str]
    means: np.ndarray[tuple[int,], np.dtype[np.float64]]
    cov: np.ndarray[tuple[int, int], np.dtype[np.float64]] | None

    def metric_data(self, metric_name: str) -> dict[str, float | None]:
        """Construct the 'Metric data-related columns' as defined in ax.Data.

        This consists of:
            - "mean": mean estimate of this parameter
            - "sem": as defined `here <https://ax.dev/docs/glossary.html#sem>`_
        """
        if metric_name not in self.metric_names:
            msg = f"Metric name {metric_name} not in the results (metrics available for {self.metric_names})"
            raise KeyError(msg)

        idx = self.metric_names.index(metric_name)

        return {"mean": self.means[idx], "sem": np.sqrt(self.cov[idx][idx]) if self.cov is not None else None}


# TODO(sw): is there any meaningful difference between closure and partial here?
# TODO(ks): Make Protocol that is more general for EvaluationFunction and make the class below an instance of that


class EvaluationFunction:
    """Class for evaluating a simulation at a point, and fits a distribution to the results.

    Take ax.Parameter, runs the simulator at that point, and packages the result into the format later required
    for unpacking in the metrics.


    """

    def __init__(
        self,
        simulator: Simulator,
        output_dist: rv_continuous,
        parameter_order: list[str],
        n_simulations_per_point: int = 20,
    ) -> None:
        """Initializes the EvaluationFunction.

        Args:
            simulator: Conforming to the simulator in
            output_dist: Distribution that should be used to fit the response. This is the distribution of the noise
             in the simulation, at a given x,
            n_simulations_per_point: The number of simulations to run at each point.
            parameter_order: Order of features that the simulator expects.

                .. Note::
                    This is a temporary measure to check that the order isn't ever shuffled.

                .. Todo::
                    Got through ax and check is order is always respected, so we can remove this
        """
        self.simulator = simulator
        self.output_dist = output_dist
        self.parameter_order = parameter_order
        self.n_simulations_per_point = n_simulations_per_point

    def __call__(self, parameters: TParameterization) -> SimulationPointResults:
        """Gets the simulation results for an input point."""
        # Not sure if TParameterization always return a dict with parameters in the same order
        # This is small risk, but will have a big impact and be difficult to catch,
        # check it here until we are completely sure
        assert list(parameters.keys()) == self.parameter_order

        x = np.array([np.fromiter(parameters.values(), dtype=float)])

        y = self.run_simulator(x)

        # This only works for a single output dimension for now
        assert y.shape[0] == 1, f"simulation_result.shape[0] must be 1, got: {y.shape[0]}"

        y_flat = y.flatten()

        return self.post_process_simulation_output(y_flat)

    def run_simulator(
        self, x: np.ndarray[tuple[int, int], np.dtype[np.float64]]
    ) -> np.ndarray[tuple[int, int, int], np.dtype[np.float64]]:
        """Runs the simulator at a point and returns the results.

        Args:
            x: The points at which to run the simulator with shape (n_points, n_input_dims)

        Returns:
            The results of the simulator with shape (n_points, n_simulations_per_point, n_output_dims)

        Raises:
            AssertionError: If the shape of the output of self.simulator is not as expected.
        """
        y = self.simulator(x=x, n_simulations_per_point=self.n_simulations_per_point)

        assert y.ndim == 3, f"simulation_result.ndim must be 3, got: {y.ndim}"  # noqa: PLR2004
        assert (
            y.shape[1] == self.n_simulations_per_point
        ), f"simulation_result.shape[1] must be {self.n_simulations_per_point}, got: {y.shape[1]}"
        assert y.shape[2] == 1, f"simulation_result.shape[2] currenlty only support single output, got: {y.shape[2]}"

        return y

    def post_process_simulation_output(
        self, y: np.ndarray[tuple[int, int, int], np.dtype[np.float64]]
    ) -> SimulationPointResults:
        """Post process the simulation output be fitting a the distribution to the results.

        Args:
            y: The output of the simulation with shape (n_simulations_per_points,)

                .. Note::
                    This is the output of the simulator, at a single point and a single output dimension.

        Returns:
            A SimulationPointResults object with the mean and covariance of the fitted distribution.
        """
        # NOTE: the order/index matches that returned by rv_continous.fit
        # This can be explicity collected by using distribution_parameter_names_from_scipy(rv_continous)
        param_means, cov = distibution_helpers.fit_dist_with_uncertainty(y, self.output_dist)

        # TODO(sw): Will need to change when we look at correlated output becuase will need to look at the covariances
        # Will address this once we know what the correlated GP requires as input and output
        param_names = distibution_helpers.distribution_parameter_names_from_scipy(self.output_dist)

        return SimulationPointResults(metric_names=param_names, means=param_means, cov=cov)



"""Helpers for working with the Ax Experiment class."""

import pandas as pd
from ax import Experiment


def input_out_df_from_experiment(exp: Experiment) -> pd.DataFrame:
    """Shows the x and y data in one dataframe."""
    df_y = exp.fetch_data().df
    df_x = pd.DataFrame([{**trial.arm.parameters, "arm_name": trial.arm.name} for trial in exp.trials.values()])  # pyright: ignore[reportAttributeAccessIssue]
    return df_y.merge(df_x, on="arm_name", how="inner")
